package ru.dbotthepony.kstarbound.world.entities.projectile

import ru.dbotthepony.kbox2d.api.ContactImpulse
import ru.dbotthepony.kbox2d.api.FixtureDef
import ru.dbotthepony.kbox2d.api.Manifold
import ru.dbotthepony.kbox2d.collision.shapes.PolygonShape
import ru.dbotthepony.kbox2d.dynamics.contact.AbstractContact
import ru.dbotthepony.kstarbound.defs.projectile.ConfiguredProjectile
import ru.dbotthepony.kstarbound.defs.projectile.ProjectilePhysics
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.entities.AliveEntity
import ru.dbotthepony.kstarbound.world.entities.LogicalMovementController
import ru.dbotthepony.kstarbound.world.entities.MovementController
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import kotlin.math.PI

abstract class AbstractProjectileMovementController(entity: Projectile, val def: ConfiguredProjectile) : MovementController<Projectile>(entity) {
	var bounces = 0
		protected set

	override fun beginContact(contact: AbstractContact) {
		val dataA = contact.fixtureA.body!!.userData
		val dataB = contact.fixtureB.body!!.userData

		if (dataA is Chunk<*, *>.TileLayer || dataB is Chunk<*, *>.TileLayer) {
			bounces++

			if (def.bounces > 0 && bounces >= def.bounces) {
				// We can't detonate inside physics simulation
				entity.markForDetonation()
			}
		} else if (dataA is MovementController<*>) {
			entity.collideWithEntity(dataA.entity)
		} else if (dataB is MovementController<*>) {
			entity.collideWithEntity(dataB.entity)
		}
	}

	override fun endContact(contact: AbstractContact) {

	}

	override fun preSolve(contact: AbstractContact, oldManifold: Manifold) {

	}

	override fun postSolve(contact: AbstractContact, impulse: ContactImpulse) {

	}

	/**
	 * Applies linear velocity along current facing angle scaled with [ConfiguredProjectile.speed]
	 */
	open fun push() {
		body.linearVelocity += Vector2d.POSITIVE_Y.rotate(body.angle) * def.speed
	}

	protected open fun updateAngle() {
		body.setTransform(position, body.linearVelocity.normalized.toAngle())
	}

	override fun think(delta: Double) {
		super.think(delta)
		updateAngle()
	}

	companion object {
		fun factorize(entity: Projectile, def: ConfiguredProjectile): MovementController<*>? {
			return when (def.physics) {
				ProjectilePhysics.DEFAULT -> LogicalMovementController(entity)
				ProjectilePhysics.BOUNCY -> BouncyPhysics(entity, def)
				ProjectilePhysics.FLAME -> FlamePhysics(entity, def)
				else -> null
			}
		}
	}
}

class BouncyPhysics(entity: Projectile, def: ConfiguredProjectile) : AbstractProjectileMovementController(entity, def) {
	override fun onSpawnedInWorld() {
		super.onSpawnedInWorld()

		body.createFixture(FixtureDef(
			shape = PolygonShape().also { it.setAsBox(0.5, 0.2) },
			restitution = 0.9,
			friction = 0.7,
			density = 2.0,
		))
	}
}

class FlamePhysics(entity: Projectile, def: ConfiguredProjectile) : AbstractProjectileMovementController(entity, def) {
	override fun onSpawnedInWorld() {
		super.onSpawnedInWorld()

		body.createFixture(FixtureDef(
			shape = PolygonShape().also { it.setAsBox(0.2, 0.2) },
			restitution = 0.0,
			friction = 1.0,
			density = 0.3,
		))
	}

	private var touchedGround = false
	private var fixedRotation = false

	override fun updateAngle() {
		if (!fixedRotation && !touchedGround)
			super.updateAngle()
	}

	override fun think(delta: Double) {
		super.think(delta)

		if (touchedGround && !fixedRotation) {
			fixedRotation = true
			body.setTransform(body.position, -PI / 2.0)
			body.isFixedRotation = true
		}
	}

	override fun beginContact(contact: AbstractContact) {
		super.beginContact(contact)
		touchedGround = true
	}
}

